"""Home Assistant Button."""

from typing import Any

from .common import Base


class Button(Base):
    """Home Assistant Button."""

    command_topic: str | None
    command_template: str | None
    payload_press: str | None

    def __init__(
        self,
        command_topic: str | None = None,
        command_template: str | None = None,
        payload_press: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise Home Assistant button."""
        super().__init__(**kwargs)
        self.command_topic = command_topic
        self.command_template = command_template
        self.payload_press = payload_press



"""Home Assistant Binary Sensor."""

from typing import Any

from .common import Base


class BinarySensor(Base):
    """Home Assistant Binary Sensor."""

    payload_on: str | None
    payload_off: str | None

    def __init__(
        self,
        payload_on: str | None = None,
        payload_off: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise Home Assistant Binary Sensor."""
        super().__init__(**kwargs)
        self.payload_on = payload_on
        self.payload_off = payload_off



"""Utils."""


from typing import Any


def dict_merge(dct: Any, merge_dct: dict[str, Any]) -> None:
    """Recursive dict merge.

    Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None.
    """
    for k in merge_dct:
        if k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], dict):
            dict_merge(dct[k], merge_dct[k])
        else:
            dct[k] = merge_dct[k]



"""Home Assistant Sensor."""


from typing import Any

from .common import Base


class Sensor(Base):
    """Home Assistant Sensor."""

    unit_of_measurement: str | None
    suggested_display_precision: int | None
    state_class: str | None

    def __init__(
        self,
        unit_of_measurement: str | None = None,
        suggested_display_precision: int | None = None,
        state_class: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise Home Assistant sensor."""
        super().__init__(**kwargs)
        self.unit_of_measurement = unit_of_measurement
        self.suggested_display_precision = suggested_display_precision
        self.state_class = state_class



"""Home Assistant Number."""

from typing import Any

from .common import Base


class Number(Base):
    """Home Assistant Number."""

    command_topic: str | None
    command_template: str | None
    min: float | None  # noqa: A003
    max: float | None  # noqa: A003
    mode: str | None
    optimistic: bool | None
    unit_of_measurement: str | None

    def __init__(  # noqa: PLR0913
        self,
        state_topic: str | None = None,
        command_topic: str | None = None,
        command_template: str | None = None,
        min: float | None = None,  # noqa: A002
        max: float | None = None,  # noqa: A002
        mode: str | None = None,
        step: float | None = None,
        optimistic: bool | None = None,
        unit_of_measurement: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialise Home Assistant number."""
        super().__init__(**kwargs)
        self.state_topic = state_topic
        self.command_topic = command_topic
        self.command_template = command_template
        self.min = min
        self.max = max
        self.mode = mode
        self.step = step
        self.optimistic = optimistic
        self.unit_of_measurement = unit_of_measurement



"""Repeat Timer module."""

from threading import Timer
from typing import Any


class RepeatTimer:
    """Repeat Timer."""

    _timer: Timer

    interval: float
    function: Any
    args: Any
    kwargs: Any
    is_running: bool

    def __init__(
        self,
        interval: float,
        function: Any,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialise Repeat Timer object."""
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False

    def _run(self) -> None:
        """Run timer."""
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self) -> None:
        """Start the thread timer."""
        if not self.is_running:
            self._timer = Timer(self.interval, self._run)
            self._timer.daemon = True
            self._timer.start()
            self.is_running = True

    def stop(self) -> None:
        """Stop the thread timer."""
        self._timer.cancel()
        self.is_running = False



"""MQTT module.

Provides the connection and events around MQTT.
"""
import logging
import socket
import sys
from typing import Any

from paho.mqtt.client import Client, MQTTMessage, MQTTv5, MQTTv311

from redreactor.helpers.emitter import EventEmitter


class MQTT:
    """MQTT module."""

    logger = logging.getLogger("Red Reactor")

    event: EventEmitter = EventEmitter()

    # MQTT Client
    _client: Client

    # Static file configuration
    _static_configuration: dict[str, Any] = {}

    # Exit on fail (stops if it cannot connect to the MQTT Broker)
    _exit_on_fail: bool = False

    def __init__(
        self,
        static_configuration: Any,
        exit_on_fail: bool = False,  # noqa: FBT001, FBT002
    ) -> None:
        """Initialise MQTT object."""
        self._static_configuration = static_configuration
        self._exit_on_fail = exit_on_fail

        self.event.on(event_name="publish", function=self.mqtt_publish)

        self.logger.info(
            "Configuring MQTT connection to: %s:%i",
            static_configuration["mqtt"]["broker"],
            static_configuration["mqtt"]["port"],
        )
        self.logger.debug(
            "Using MQTT Version %s, Protocol: %s",
            static_configuration["mqtt"]["version"],
            static_configuration["mqtt"]["transport"],
        )
        self.logger.debug(
            "MQTT `Exit on Fail` is %s",
            ("disabled", "enabled")[self._exit_on_fail],
        )

        try:
            # Check to see if the MQTT Version being used is MQTT V5
            if int(static_configuration["mqtt"]["version"]) == 5:
                self._client = Client(
                    client_id=static_configuration["mqtt"]["client_id"],
                    protocol=MQTTv5,
                    transport=static_configuration["mqtt"]["transport"],
                )

            # Check to see if the MQTT Version being used is MQTT V3.11
            if int(static_configuration["mqtt"]["version"]) == 3:
                self._client = Client(
                    client_id=static_configuration["mqtt"]["client_id"],
                    protocol=MQTTv311,
                    clean_session=True,
                    transport=static_configuration["mqtt"]["transport"],
                )
        except (ValueError, AttributeError):
            self.logger.exception("MQTT Client Error")

            if self._exit_on_fail:
                sys.exit()
        except NameError:
            self.logger.exception(
                "MQTT Client Version number is invalid, must be `3` or `5`",
            )

            if self._exit_on_fail:
                sys.exit(0)

        self._client.on_connect = self._mqtt_on_connect
        self._client.on_disconnect = self._mqtt_on_disconnect
        self._client.on_message = self._mqtt_on_message

        self._client.username_pw_set(
            static_configuration["mqtt"]["user"],
            static_configuration["mqtt"]["password"],
        )

    def connect(
        self,
        broker: str = "127.0.0.1",
        port: int = 1883,
        timeout: int = 60,
    ) -> None:
        """MQTT Connect to the Broker."""
        self.logger.debug("Initiating connection to Broker: %s:%i", broker, port)

        try:
            self._client.connect(host=broker, port=port, keepalive=timeout)
        except (ConnectionRefusedError, socket.timeout):
            self.logger.exception("Unable to connect to the MQTT Broker")

            if self._exit_on_fail:
                sys.exit(0)

        # Run forever until battery shutdown or user exit
        self._client.loop_forever()

    def _mqtt_on_connect(  # noqa: PLR0913
        self,
        client: Client,
        userdata: Any,
        flags: Any,
        rc: Any,
        reasoncode: Any = None,
    ) -> None:
        """MQTT Connect event."""
        if rc == 0:
            self.logger.info("Connected to MQTT Broker")

            # Configure a last will message if something happens to the connection
            client.will_set(
                topic=f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['status']}",
                payload=f"{self._static_configuration['status']['offline']}",
                qos=1,
                retain=True,
            )

            # Notify MQTT event on_connect listeners
            self.event.emit(
                event_name="on_connect",
                client=client,
                userdata=userdata,
                flags=flags,
                rc=rc,
                reasoncode=reasoncode,
            )
        else:
            self.logger.error("MQTT Broker connection refused, error code %i", rc)

    def _mqtt_on_disconnect(
        self,
        client: Client,
        userdata: Any,
        rc: Any,
        reasoncode: Any = None,
    ) -> None:
        """MQTT Disconnect event."""
        self.logger.debug("Notifying listeners of client disconnect")

        self.event.emit(
            event_name="on_disconnect",
            client=client,
            userdata=userdata,
            rc=rc,
            reasoncode=reasoncode,
        )

        self.logger.info("Disconnected from broker")

    def _mqtt_on_message(
        self,
        client_id: str,
        userdata: Any,
        message: MQTTMessage,
    ) -> None:
        """MQTT On Message event."""
        self.logger.debug(
            "Message received: %s - %s",
            str(message.topic),
            str(message.payload.decode("utf-8")),
        )

        self.event.emit(
            event_name="on_message",
            client_id=client_id,
            userdata=userdata,
            message=message,
        )

    def mqtt_publish(self, topic: str, payload: str) -> None:
        """Publish data to a MQTT Topic.

        Used to publish data on a specified MQTT topic, with a specified payload.
        """
        self.logger.debug("Publishing: %s - %s", topic, payload)

        self._client.publish(topic=topic, payload=payload)



"""Monitor Data file."""


from redreactor.const import (
    DEFAULT_BATTERY_VOLTAGE_MAXIMUM,
    DEFAULT_BATTERY_VOLTAGE_MINIMUM,
    DEFAULT_BATTERY_WARNING_THRESHOLD,
    DEFAULT_REPORT_INTERVAL,
)


class MonitorData:
    """Monitor Data."""

    voltage: float
    current: float
    battery_level: float
    external_power: bool
    cpu_temperature: float | None
    cpu_stat: int | None
    battery_warning_threshold: int
    battery_voltage_minimum: float
    battery_voltage_maximum: float
    report_interval: int

    def __init__(  # noqa: PLR0913
        self,
        voltage: float = 0.0,
        current: float = 0.0,
        battery_level: float = 0.0,
        external_power: bool = True,  # noqa: FBT001, FBT002
        cpu_temperature: float = 0.0,
        cpu_stat: int = 0,
        battery_warning_threshold: int = DEFAULT_BATTERY_WARNING_THRESHOLD,
        battery_voltage_minimum: float = DEFAULT_BATTERY_VOLTAGE_MINIMUM,
        battery_voltage_maximum: float = DEFAULT_BATTERY_VOLTAGE_MAXIMUM,
        report_interval: int = DEFAULT_REPORT_INTERVAL,
    ) -> None:
        """Initialise Monitor Data."""
        self.voltage = voltage
        self.current = current
        self.battery_level = battery_level
        self.external_power = external_power
        self.cpu_temperature = cpu_temperature
        self.cpu_stat = cpu_stat
        self.battery_warning_threshold = battery_warning_threshold
        self.battery_voltage_minimum = battery_voltage_minimum
        self.battery_voltage_maximum = battery_voltage_maximum
        self.report_interval = report_interval



"""Configuration module.

Contains the ability to read the static and dynamic configuration files.
"""
import json
import os
from typing import Any

import yaml

from redreactor.const import (
    DEFAULT_BATTERY_VOLTAGE_MAXIMUM,
    DEFAULT_BATTERY_VOLTAGE_MINIMUM,
    DEFAULT_BATTERY_WARNING_THRESHOLD,
    DEFAULT_INA_I2C_ADDRESS,
    DEFAULT_INA_MAX_EXPECTED_AMPS,
    DEFAULT_INA_MONITOR_INTERVAL,
    DEFAULT_INA_SHUNT_OHMS,
    DEFAULT_REPORT_INTERVAL,
)
from redreactor.helpers.emitter import EventEmitter
from redreactor.helpers.utils import dict_merge


class DynamicConfiguration:
    """Dynamic Configuration object."""

    data: dict[str, Any]

    event: EventEmitter = EventEmitter()

    _dynamic_file: str

    def __init__(self, dynamic_file: str) -> None:
        """Initialise the Dynamic Configuration object."""
        self._dynamic_file = dynamic_file

        self.data = self._load()

        self.event.on(event_name="write", function=self.write)

    def _load(self) -> Any:
        """Load dynamic configuration file."""
        dynamic_file_defaults = {
            "report_interval": DEFAULT_REPORT_INTERVAL,
            "battery_warning_threshold": DEFAULT_BATTERY_WARNING_THRESHOLD,
            "battery_voltage_minimum": DEFAULT_BATTERY_VOLTAGE_MINIMUM,
            "battery_voltage_maximum": DEFAULT_BATTERY_VOLTAGE_MAXIMUM,
        }

        if not os.path.exists(self._dynamic_file):  # noqa: PTH110
            with open(self._dynamic_file, "w") as file:  # noqa: PTH123
                json.dump(dynamic_file_defaults, file)

        with open(self._dynamic_file) as file:  # noqa: PTH123
            dynamic_file_override = json.load(file)

        return {**dynamic_file_defaults, **dynamic_file_override}

    def write(self) -> None:
        """Write dynamic configuration file."""
        with open(self._dynamic_file, "w") as file:  # noqa: PTH123
            json.dump(self.data, file)


class LinkedConfiguration:
    """Linked Configuration.

    A connection between both Dynamic and Static configurations.
    """

    static: dict[str, Any]
    dynamic: DynamicConfiguration

    _static_file: str
    _dynamic_file: str

    def __init__(self, static_file: str, dynamic_file: str) -> None:
        """Initialise Linked Configuration object."""
        self._static_file = static_file
        self._dynamic_file = dynamic_file

        self.static = self._load()
        self.dynamic = DynamicConfiguration(self._dynamic_file)

    def _load(self) -> dict[str, Any]:
        """Load static configuration file, and override defaults."""
        with open(self._static_file) as static_file:  # noqa: PTH123
            static_file_override = yaml.safe_load(static_file)

        static_file_defaults = {
            "mqtt": {
                "broker": "127.0.0.1",
                "port": 1883,
                "user": "test",
                "password": "test",
                "client_id": "Red Reactor",
                "base_topic": "redreactor",
                "version": 3,
                "transport": "tcp",
                "timeout": 120,
                "topic": {
                    "state": "state",
                    "status": "status",
                    "set": "set",
                },
                "exit_on_fail": True,
            },
            "hostname": {"name": "redreactor", "pretty": "Red Reactor"},
            "homeassistant": {
                "discovery": True,
                "topic": "homeassistant",
                "discovery_interval": 120,
                "expire_after": 120,
            },
            "status": {
                "online": "online",
                "offline": "offline",
            },
            "fields": {
                "voltage": {
                    "name": "voltage",
                    "pretty": "Voltage",
                    "type": "sensor",
                    "unit": "V",
                    "device_class": "voltage",
                    "suggested_display_precision": 2,
                },
                "current": {
                    "name": "current",
                    "pretty": "Current",
                    "type": "sensor",
                    "unit": "mA",
                    "device_class": "current",
                    "suggested_display_precision": 2,
                },
                "battery_level": {
                    "name": "battery_level",
                    "pretty": "Battery Level",
                    "type": "sensor",
                    "unit": "%",
                    "device_class": "battery",
                },
                "external_power": {
                    "name": "external_power",
                    "pretty": "External Power",
                    "type": "binary_sensor",
                    "device_class": "plug",
                    "entity_category": "diagnostic",
                },
                "cpu_temperature": {
                    "name": "cpu_temperature",
                    "pretty": "CPU Temperature",
                    "type": "sensor",
                    "unit": "°C",
                    "device_class": "temperature",
                    "entity_category": "diagnostic",
                    "suggested_display_precision": 2,
                },
                "cpu_stat": {
                    "name": "cpu_stat",
                    "pretty": "CPU Stat",
                    "type": "sensor",
                    "entity_category": "diagnostic",
                },
                "battery_warning_threshold": {
                    "name": "battery_warning_threshold",
                    "pretty": "Battery Warning",
                    "type": "number",
                    "unit": "%",
                    "device_class": "battery",
                    "entity_category": "diagnostic",
                    "min": 0,
                    "max": 100,
                    "mode": "box",
                    "step": 1,
                },
                "battery_voltage_minimum": {
                    "name": "battery_voltage_minimum",
                    "pretty": "Battery Voltage Minimum",
                    "type": "number",
                    "unit": "V",
                    "device_class": "voltage",
                    "entity_category": "diagnostic",
                    "min": 2.5,
                    "max": 4.5,
                    "mode": "box",
                    "step": 0.1,
                },
                "battery_voltage_maximum": {
                    "name": "battery_voltage_maximum",
                    "pretty": "Battery Voltage Maximum",
                    "type": "number",
                    "unit": "V",
                    "device_class": "voltage",
                    "entity_category": "diagnostic",
                    "min": 2.5,
                    "max": 4.5,
                    "mode": "box",
                    "step": 0.1,
                },
                "report_interval": {
                    "name": "report_interval",
                    "pretty": "Report Interval",
                    "type": "number",
                    "unit": "s",
                    "entity_category": "diagnostic",
                    "min": 5,
                    "max": 300,
                    "mode": "box",
                    "step": 5,
                },
                "restart": {
                    "name": "restart",
                    "pretty": "Restart",
                    "type": "button",
                    "entity_category": "diagnostic",
                },
                "shutdown": {
                    "name": "shutdown",
                    "pretty": "Shutdown",
                    "type": "button",
                    "entity_category": "diagnostic",
                },
            },
            "ina": {
                "address": DEFAULT_INA_I2C_ADDRESS,
                "shunt_ohms": DEFAULT_INA_SHUNT_OHMS,
                "max_expected_amps": DEFAULT_INA_MAX_EXPECTED_AMPS,
                "monitor_interval": DEFAULT_INA_MONITOR_INTERVAL,
            },
            "system": {
                "shutdown": "sudo shutdown 0 -h",
                "restart": "sudo shutdown 0 -r",
            },
            "logging": {
                "console": "INFO",
                "file": "WARNING",
            },
        }

        # Overwritten any different values with the ones that came from the static file
        dict_merge(static_file_defaults, static_file_override)
        return static_file_defaults



"""Commander module.

Adds support for receiving and processing MQTT commands.
"""
import json
import logging
import os
import sys
import time
from datetime import datetime
from typing import Any, Literal

from paho.mqtt.client import Client, MQTTMessage

from redreactor.components.monitor import Monitor
from redreactor.components.mqtt import MQTT
from redreactor.configuration import DynamicConfiguration
from redreactor.helpers.emitter import EventEmitter


class Commander:
    """Commander for processing received MQTT commands."""

    logger = logging.getLogger("Red Reactor")

    event: EventEmitter = EventEmitter()

    _static_configuration: dict[str, Any] = {}
    _dynamic_configuration: DynamicConfiguration
    _monitor: Monitor

    def __init__(
        self,
        static_configuration: dict[str, Any],
        dynamic_configuration: DynamicConfiguration,
        monitor: Monitor,
    ) -> None:
        """Initialise the Commander object."""
        self._static_configuration = static_configuration
        self._dynamic_configuration = dynamic_configuration
        self._monitor = monitor

        self.logger.debug("Setup the command handler")

        # Register Shutdown event listener
        self._monitor.event.on(event_name="shutdown", function=self._shutdown)

        # Register MQTT events
        MQTT.event.on(event_name="on_connect", function=self._on_connect)
        MQTT.event.on(event_name="on_message", function=self._on_message)

    def _on_connect(  # noqa: PLR0913
        self,
        client: Client,
        userdata: Any,  # noqa: ARG002
        flags: Any,  # noqa: ARG002
        rc: Any,  # noqa: ARG002
        reasoncode: Any = None,  # noqa: ARG002
    ) -> None:
        """Process Commander subscriptions on MQTT connect event."""
        # Loop through all available field options
        for field in self._static_configuration["fields"]:
            # Only process fields that are of type button or number
            if self._static_configuration["fields"][field]["type"] in {
                "number",
                "button",
            }:
                self.logger.info(
                    "Subscribing to the %s topic at: %s/%s/%s/%s",
                    self._static_configuration["fields"][field]["pretty"],
                    self._static_configuration["mqtt"]["base_topic"],
                    self._static_configuration["hostname"]["name"],
                    self._static_configuration["mqtt"]["topic"]["set"],
                    self._static_configuration["fields"][field]["name"],
                )

                # Subscribe to that specific topic
                client.subscribe(
                    topic=f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['set']}/{self._static_configuration['fields'][field]['name']}",
                )

    def _on_message(
        self,
        client_id: str,  # noqa: ARG002
        userdata: Any,  # noqa: ARG002
        message: MQTTMessage,
    ) -> None:
        """Process Commander messages on MQTT message event."""
        # Remove the start of the topic to leave the command
        topic = str(message.topic).replace(
            f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['set']}/",
            "",
        )

        # Convert the received command into readable JSON
        try:
            command = json.loads(str(message.payload.decode("utf-8")))
        except json.JSONDecodeError:
            self.logger.warning(
                "JSON Decode Error: %s",
                str(message.payload.decode("utf-8")),
            )
            return

        # Loop through available fields to find the expected command topic
        for field in self._static_configuration["fields"]:
            # If the received command is of type button
            if (
                self._static_configuration["fields"][field]["type"]
                in {
                    "button",
                }
                and field in topic
            ):
                self.logger.info(
                    "%s button has been pressed",
                    self._static_configuration["fields"][field]["pretty"],
                )
                self._on_command(event_type=field)
                return

            # If the received command is of type number
            if (
                self._static_configuration["fields"][field]["type"]
                in {
                    "number",
                }
                and field in topic
            ):
                self.logger.info(
                    "%s update command received, changing to value %s",
                    self._static_configuration["fields"][field]["pretty"],
                    command,
                )

                # Update Dynamic configuration with latest value
                self._dynamic_configuration.data[field] = command

                # Dynamic configuration has changed, call write event
                self._dynamic_configuration.event.emit("write")
                return

        # If it gets to here, the received command is invalid
        self.logger.warning("Command received on %s is invalid", topic)

    def _on_command(self, event_type: Literal["shutdown", "restart"]) -> None:
        """Commander's on command event.

        Must contain an event_type of 'shutdown' or 'restart'.
        """
        # Check to ensure that the event_type is either shutdown or restart
        if event_type not in {"shutdown", "restart"}:
            msg = "Invalid device command."
            raise ValueError(msg)

        self.logger.info(
            "Device %s has been called. Last running at %s",
            event_type,
            datetime.now().strftime("%H:%M:%S"),  # noqa: DTZ005
        )

        # Publish to MQTT the offline state, before shutting down or restarting
        MQTT.event.emit(
            event_name="publish",
            topic=f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['status']}",
            payload=f"{self._static_configuration['status']['offline']}",
        )

        # Ensure the MQTT data has been published to the broker
        time.sleep(2)

        # Initiate event type (shutdown or restart)
        os.system(self._static_configuration["system"][event_type])  # noqa: S605

        # Exit the program
        sys.exit(0)

    def _shutdown(self) -> None:
        """Shutdown event hook."""
        self._on_command(event_type="shutdown")



"""Event Emitter.

Custom event emitter.
"""
from typing import Any


class EventEmitter:
    """Event Emitter module.

    Allows the creation of callback style functionality.
    """

    def __init__(self) -> None:
        """Initialise Event Emitter."""
        self.__callbacks: Any = {}

    def on(self, event_name: str, function: Any) -> Any:
        """Register event to a specific item."""
        self.__callbacks[event_name] = [*self.__callbacks.get(event_name, []), function]
        return function

    def emit(
        self,
        event_name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Event trigger."""
        for function in self.__callbacks.get(event_name, []):
            function(*args, **kwargs)

    def off(self, event_name: str, function: Any) -> None:
        """Event stop."""
        self.__callbacks.get(event_name, []).remove(function)



"""Home Assistant Common."""

from json import JSONEncoder
from typing import Any


# Home Assistant MQTT Representer object
# Enabled the printing into a dict and to JSON
class Representer:
    """Representer.

    Enables printing into a dict and to JSON.
    """

    def __repr__(self) -> str:
        """Return Represent."""
        return repr({k: v for (k, v) in self.__dict__.items() if v is not None})


# Home Assistant MQTT JSON Encoder
# Allows the conversion of the Home Assistant objects to JSON
class Encoder(JSONEncoder):
    """Encoder.

    Allows the conversion of the Home Assistant object into JSON.
    """

    def default(self, o: Any) -> Any:
        """Encode object into JSON."""
        if isinstance(o, Device | Availability):
            return {k: v for (k, v) in o.__dict__.items() if v is not None}

        return JSONEncoder.default(self, o)


# Home Assistant MQTT Availability topic
# Topic to subscribe to receive availability updates
class Availability(Representer):
    """Availability.

    Adds Home Assistant MQTT Availability support.
    """

    topic: str
    payload_available: str | None
    payload_not_available: str | None

    def __init__(
        self,
        topic: str,
        payload_available: str | None = None,
        payload_not_available: str | None = None,
    ) -> None:
        """Initialise Home Assistant Availability object."""
        self.payload_available = payload_available
        self.payload_not_available = payload_not_available
        self.topic = topic


# Home Assistant MQTT Device
# Ties the device into the device registry within Home Assistant
class Device(Representer):
    """Device.

    Adds support for Home Assistant device list.
    """

    configuration_url: str | None
    connections: list[list[str]] | None  # [connection_type, connection_identifier]
    hw_version: str | None
    identifiers: str | list[str] | None
    manufacturer: str | None
    model: str | None
    name: str | None
    suggested_area: str | None
    sw_version: str | None
    via_device: str | None

    def __init__(  # noqa: PLR0913
        self,
        configuration_url: str | None = None,
        connections: list[list[str]] | None = None,
        hw_version: str | None = None,
        identifiers: str | list[str] | None = None,
        manufacturer: str | None = None,
        model: str | None = None,
        name: str | None = None,
        suggested_area: str | None = None,
        sw_version: str | None = None,
        via_device: str | None = None,
    ) -> None:
        """Initialise Home Assistant Device object."""
        self.configuration_url = configuration_url
        self.connections = connections
        self.hw_version = hw_version
        self.identifiers = identifiers
        self.manufacturer = manufacturer
        self.model = model
        self.name = name
        self.suggested_area = suggested_area
        self.sw_version = sw_version
        self.via_device = via_device


# Home Assistant MQTT Items
# Standard options available on all MQTT inputs
class Base(Representer):
    """Base.

    In all Home Assistant device types.
    """

    name: str | None
    device_class: str | None
    state_class: str | None
    entity_category: str | None
    expire_after: int | None  # Number of seconds after state becomes unavailable
    icon: str | None
    object_id: str | None
    unique_id: str | None
    state_topic: str | None
    value_template: str | None
    availability: list[Availability] | None
    availability_mode: str | None
    device: Device | None

    configuration_topic: str | None  # Used for one time creation of the Home Assistant configuration topic # noqa: E501

    def __init__(  # noqa: PLR0913
        self,
        name: str | None = None,
        device_class: str | None = None,
        state_class: str | None = None,
        entity_category: str | None = None,
        expire_after: int | None = None,
        icon: str | None = None,
        object_id: str | None = None,
        unique_id: str | None = None,
        state_topic: str | None = None,
        value_template: str | None = None,
        availability: list[Availability] | None = None,
        availability_mode: str | None = None,
        device: Device | None = None,
        configuration_topic: str | None = None,
    ) -> None:
        """Initialise Home Assistant Base object."""
        self.name = name
        self.device_class = device_class
        self.state_class = state_class
        self.entity_category = entity_category
        self.expire_after = expire_after
        self.icon = icon
        self.object_id = object_id
        self.unique_id = unique_id
        self.state_topic = state_topic
        self.value_template = value_template
        self.availability = availability
        self.availability_mode = availability_mode
        self.device = device

        self.configuration_topic = configuration_topic



"""Constants used by Red Reactor components."""

DEFAULT_REPORT_INTERVAL = 30  # Seconds between data reporting
DEFAULT_EXPIRE_INTERVAL = 120  # Seconds before Home Assistant invalidates received data
DEFAULT_BATTERY_WARNING_THRESHOLD = 10  # Percentage of charge remaining
DEFAULT_BATTERY_VOLTAGE_MINIMUM = 2.9  # Minimum battery voltage before shutting down
DEFAULT_BATTERY_VOLTAGE_MAXIMUM = 4.2  # Maximum battery voltage
DEFAULT_BATTERY_VOLTAGE_MAXIMUM_DROP = 0.03  # Maximum battery voltage allowable drop
DEFAULT_BATTERY_VOLTAGE_ERROR = DEFAULT_BATTERY_VOLTAGE_MAXIMUM + 0.05

DEFAULT_INA_I2C_ADDRESS = 0x40  # Red Reactor Default I2C Address
DEFAULT_INA_SHUNT_OHMS = 0.05  # Red Reactor Shunt Ohms
DEFAULT_INA_MAX_EXPECTED_AMPS = 5.5
DEFAULT_INA_MONITOR_INTERVAL = 5  # Second read time



"""Monitor module."""

import json
import logging
import subprocess
import sys
from typing import Any

from ina219 import INA219, DeviceRangeError
from paho.mqtt.client import Client

from redreactor.components.mqtt import MQTT
from redreactor.configuration import DynamicConfiguration
from redreactor.const import (
    DEFAULT_BATTERY_VOLTAGE_MAXIMUM,
    DEFAULT_BATTERY_VOLTAGE_MAXIMUM_DROP,
    DEFAULT_BATTERY_VOLTAGE_MINIMUM,
    DEFAULT_BATTERY_WARNING_THRESHOLD,
    DEFAULT_REPORT_INTERVAL,
)
from redreactor.helpers.emitter import EventEmitter
from redreactor.helpers.repeater import RepeatTimer

from .data import MonitorData


class Monitor:
    """Monitor module.

    Used to monitor the state of the Red Reactor
    """

    logger = logging.getLogger("Red Reactor")

    event: EventEmitter = EventEmitter()

    _static_configuration: dict[str, Any]
    _dynamic_configuration: DynamicConfiguration

    # Adds default values into the Monitor Data dict
    data: MonitorData = MonitorData(
        voltage=0.0,
        current=0.0,
        battery_level=100,
        external_power=True,
        cpu_temperature=0.0,
        cpu_stat=0,
        battery_warning_threshold=DEFAULT_BATTERY_WARNING_THRESHOLD,
        battery_voltage_minimum=DEFAULT_BATTERY_VOLTAGE_MINIMUM,
        battery_voltage_maximum=DEFAULT_BATTERY_VOLTAGE_MAXIMUM,
        report_interval=DEFAULT_REPORT_INTERVAL,
    )

    # Timers
    monitor_timer: RepeatTimer  # Runs Monitoring calculations
    report_timer: RepeatTimer  # Pushes updates to MQTT

    def __init__(
        self,
        static_configuration: Any,
        dynamic_configuration: DynamicConfiguration,
    ) -> None:
        """Initialise Monitor object."""
        self._static_configuration = static_configuration
        self._dynamic_configuration = dynamic_configuration

        # Register Dynamic Configuration update listener
        self._dynamic_configuration.event.on(event_name="write", function=self._update)

        # Register MQTT on connect listener
        MQTT.event.on(event_name="on_connect", function=self._mqtt_on_connect)

        self.logger.info("Initiating battery monitor")

        # Configure the report timer, gets started on MQTT connect
        self.report_timer = RepeatTimer(
            self.data.report_interval,
            self._update,
        )

        # Update Dynamic Configuration values to there latest values
        self._update_dynamic_configuration()

        # Setup the INA219
        ina: INA219 | None = None
        try:
            ina = INA219(
                shunt_ohms=static_configuration["ina"]["shunt_ohms"],
                max_expected_amps=static_configuration["ina"]["max_expected_amps"],
                address=static_configuration["ina"]["address"],
                busnum=1,
                log_level=logging.ERROR,
            )
            ina.configure(voltage_range=ina.RANGE_16V)
        except (OSError, ModuleNotFoundError):
            self.logger.exception("Unable to connect to the Red Reactor")

            # Exit the program
            sys.exit(0)

        # Configure the monitor timer, start's as soon as possible
        self.monitor_timer = RepeatTimer(
            float(self._static_configuration["ina"]["monitor_interval"]),
            self._monitor,
            ina=ina,
        )
        self.monitor_timer.start()

    def _monitor(self, ina: INA219) -> None:  # noqa: PLR0912
        """Monitor INA Thread Loop.

        Gets called every 'monitor_interval'.
        """
        self.logger.debug("Monitoring thread has been started")

        # Update the Dynamic Configuration to ensure the calculations will be based correctly off the changeable values # noqa: E501
        self._update_dynamic_configuration()

        shutdown: bool = False

        if ina:
            try:
                self.data.voltage = ina.voltage()
                self.data.battery_level = self._calculate_battery_level(
                    self.data.voltage,
                )
                self.data.current = ina.current()
            except DeviceRangeError:
                # Assume no external power so the device will still shutdown on a low voltage reading # noqa: E501
                self.logger.exception("Battery current range error")
                self.data.external_power = False
                self.data.current = 6000

                MQTT.event.emit(
                    event_name="publish",
                    topic=f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['status']}",  # noqa: E501
                    payload=f"{self._static_configuration['status']['offline']}",
                )
            except ZeroDivisionError:
                self.logger.exception(
                    "Minimum Voltage %fV and Maximum Voltage %fV difference is equal to 0",  # noqa: E501
                    self.data.battery_voltage_minimum,
                    self.data.battery_voltage_maximum,
                )
            else:
                # Identify status change
                if self.data.current > 10:
                    # No External Power
                    if self.data.external_power:
                        # Power removed
                        self.data.external_power = False
                        self._update()
                elif self.data.current >= 0:
                    # Battery now Full
                    self.data.external_power = True
                elif not self.data.external_power:
                    # Power restored
                    self.data.external_power = True
                    self._update()

        if (
            self.data.battery_level <= float(self.data.battery_warning_threshold)
            and not self.data.external_power
        ):
            # Force immediate publish update at warning level
            self._update()

        if self.data.battery_level == 0.0 and not self.data.external_power:
            shutdown = True

        if self.data.voltage > self.data.battery_voltage_maximum + 0.05:
            self.data.external_power = True
            # Force immediate publish update at warning level
            self._update()

        # Shutdown system
        if shutdown:
            self.logger.warning(
                "Forcing system shutdown, going offline with %fV",
                round(self.data.voltage, 2),
            )

            # Emit the shutdown event
            self.event.emit("shutdown")

    def _calculate_battery_level(self, voltage: float) -> int:
        """Calculate Battery Level."""
        return int(
            max(
                min(
                    100,
                    (voltage - self.data.battery_voltage_minimum)
                    / (
                        self.data.battery_voltage_maximum
                        - DEFAULT_BATTERY_VOLTAGE_MAXIMUM_DROP
                        - self.data.battery_voltage_minimum
                    )
                    * 100,
                ),
                0,
            ),
        )

    def _update(self) -> None:
        """Pushes the latest data to Monitor endpoint."""
        # Force a update of the dynamic configuration values
        self._update_dynamic_configuration()

        # Read data from Kernel endpoints
        try:
            cpu_temperature = subprocess.Popen(
                ["cat", "/sys/class/thermal/thermal_zone0/temp"],  # noqa: S603, S607
                stdout=subprocess.PIPE,
            )
            result_cpu_temperature = cpu_temperature.communicate()
            self.data.cpu_temperature = round(
                float((result_cpu_temperature[0]).decode()) * 0.001,
                2,
            )

            cpu_stat = subprocess.Popen(
                [  # noqa: S603, S607
                    "cat",
                    "/sys/devices/platform/soc/soc:firmware/get_throttled",
                ],
                stdout=subprocess.PIPE,
            )
            result_cpu_stat = cpu_stat.communicate()
            self.data.cpu_stat = int(result_cpu_stat[0].decode())
        except (OSError, IndexError, ValueError):
            # Failed to extract info
            self.logger.exception("Failed to read CPU Information")
            self.data.cpu_stat = None
            self.data.cpu_temperature = None

        # Publish heartbeat to status endpoint
        MQTT.event.emit(
            event_name="publish",
            topic=f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['status']}",
            payload=f"{self._static_configuration['status']['online']}",
        )

        # Publish latest data to state endpoint
        MQTT.event.emit(
            event_name="publish",
            topic=f"{self._static_configuration['mqtt']['base_topic']}/{self._static_configuration['hostname']['name']}/{self._static_configuration['mqtt']['topic']['state']}",
            payload=json.dumps(
                {
                    f"{self._static_configuration['fields']['voltage']['name']}": round(
                        self.data.voltage,
                        3,
                    ),
                    f"{self._static_configuration['fields']['current']['name']}": round(
                        self.data.current,
                        4,
                    ),
                    f"{self._static_configuration['fields']['battery_level']['name']}": self.data.battery_level,  # noqa: E501
                    f"{self._static_configuration['fields']['external_power']['name']}": self._static_configuration[  # noqa: E501
                        "fields"
                    ][
                        "external_power"
                    ].get(
                        "payload_on",
                        "ON",
                    )
                    if self.data.external_power
                    else self._static_configuration["fields"]["external_power"].get(
                        "payload_off",
                        "OFF",
                    ),
                    f"{self._static_configuration['fields']['cpu_temperature']['name']}": self.data.cpu_temperature,  # noqa: E501
                    f"{self._static_configuration['fields']['cpu_stat']['name']}": self.data.cpu_stat,  # noqa: E501
                    f"{self._static_configuration['fields']['battery_warning_threshold']['name']}": self.data.battery_warning_threshold,  # noqa: E501
                    f"{self._static_configuration['fields']['battery_voltage_minimum']['name']}": self.data.battery_voltage_minimum,  # noqa: E501
                    f"{self._static_configuration['fields']['battery_voltage_maximum']['name']}": self.data.battery_voltage_maximum,  # noqa: E501
                    f"{self._static_configuration['fields']['report_interval']['name']}": self.data.report_interval,  # noqa: E501
                },
            ),
        )

    def _update_dynamic_configuration(self) -> None:
        """Update Dynamic Configuration.

        This uses a event hook back to the Dynamic Configuration to update the file.
        """
        # Update Battery Warning Threshold
        self.data.battery_warning_threshold = int(
            self._dynamic_configuration.data[
                self._static_configuration["fields"]["battery_warning_threshold"][
                    "name"
                ]
            ],
        )

        # Update Battery Voltage Minimum
        self.data.battery_voltage_minimum = float(
            self._dynamic_configuration.data[
                self._static_configuration["fields"]["battery_voltage_minimum"]["name"]
            ],
        )

        # Update Battery Voltage Maximum
        self.data.battery_voltage_maximum = float(
            self._dynamic_configuration.data[
                self._static_configuration["fields"]["battery_voltage_maximum"]["name"]
            ],
        )

        # Update Report Interval
        self.data.report_interval = int(
            self._dynamic_configuration.data[
                self._static_configuration["fields"]["report_interval"]["name"]
            ],
        )

        if self.report_timer:
            # Update Monitor timer interval
            self.report_timer.interval = self.data.report_interval

    def _mqtt_on_connect(  # noqa: PLR0913
        self,
        client: Client,  # noqa: ARG002
        userdata: Any,  # noqa: ARG002
        flags: Any,  # noqa: ARG002
        rc: Any,  # noqa: ARG002
        reasoncode: Any = None,  # noqa: ARG002
    ) -> None:
        """On MQTT Connect.

        Start the reporting timer.
        """
        self.report_timer.start()



"""Home Assistant module."""
import json
import logging
from typing import Any

from paho.mqtt.client import Client

from redreactor.components.mqtt import MQTT
from redreactor.configuration import DynamicConfiguration
from redreactor.helpers.repeater import RepeatTimer

from .binary_sensor import BinarySensor
from .button import Button
from .common import Availability, Base, Device, Encoder
from .number import Number
from .sensor import Sensor


class Homeassistant:
    """Home Assistant integration."""

    logger = logging.getLogger("Red Reactor")

    configuration: list[Sensor | Number | BinarySensor | Button] = []

    _static_configuration: dict[str, Any] = {}
    _dynamic_configuration: DynamicConfiguration

    configuration_report_timer: RepeatTimer

    def __init__(
        self,
        static_configuration: dict[str, Any],
        dynamic_configuration: DynamicConfiguration,
    ) -> None:
        """Initialise Home Assistant integration object."""
        self._static_configuration = static_configuration
        self._dynamic_configuration = dynamic_configuration

        self._process_homeassistant_configuration(static_configuration)

        # Register MQTT listeners
        MQTT.event.on(event_name="on_connect", function=self._mqtt_on_connect)

        # Register Dynamic Configuration update listener
        self._dynamic_configuration.event.on(
            event_name="write",
            function=self._update_homeassistant_timer,
        )

        # Push a configuration update at a repeated interval
        self.configuration_report_timer = RepeatTimer(
            float(
                self._dynamic_configuration.data[
                    self._static_configuration["fields"]["report_interval"]["name"]
                ],
            )
            * 4,
            self._update_homeassistant_configuration,
        )

        self.logger.info("Home Assistant support has been enabled")

    def _process_homeassistant_configuration(self, static_configuration: Any) -> None:
        """Process Home Assistant configuration."""
        configuration_defaults: Device = Device(
            identifiers=f"redreactor_{static_configuration['hostname']['name']}",
            name=f"Red Reactor {static_configuration['hostname']['pretty']}",
            manufacturer="Pascal Herczog",
            model="Red Reactor",
            hw_version="Rev 1.5",
            sw_version="1.0.0",
        )

        for field in static_configuration["fields"]:
            field = static_configuration["fields"][field]  # noqa: PLW2901

            configuring: Base = Base(
                name=f"{configuration_defaults.name} {field.get('pretty')}",
                device_class=field.get("device_class", None),
                state_class="measurement",
                expire_after=int(static_configuration["homeassistant"]["expire_after"]),
                entity_category=field.get("entity_category", None),
                icon=field.get("icon", None),
                object_id=field.get("object_id", None),
                unique_id=f"{configuration_defaults.identifiers}_{field['name']}",
                state_topic=f"{static_configuration['mqtt']['base_topic']}/{static_configuration['hostname']['name']}/{static_configuration['mqtt']['topic']['state']}",
                configuration_topic=f"{static_configuration['homeassistant']['topic']}/{field['type']}/{configuration_defaults.identifiers}_{field['name']}/config",
                availability=[
                    Availability(
                        topic=f"{static_configuration['mqtt']['base_topic']}/{static_configuration['hostname']['name']}/{static_configuration['mqtt']['topic']['status']}",
                        payload_available=f"{static_configuration['status']['online']}",
                        payload_not_available=f"{static_configuration['status']['offline']}",
                    ),
                ],
                availability_mode=f"{field.get('availability_mode', 'all')}",
                value_template=field.get(
                    "value_template",
                    f"{ '{{ value_json.' + str(field['name']) + ' }}' }",
                ),
                device=configuration_defaults,
            )

            configured: Sensor | BinarySensor | Number | Button = Sensor()
            if field["type"] == "sensor":
                configured = Sensor(
                    unit_of_measurement=field.get("unit", None),
                    suggested_display_precision=field.get(
                        "suggested_display_precision",
                        None,
                    ),
                    state_class=field.get("state_class", None),
                )

            if field["type"] == "binary_sensor":
                configured = BinarySensor(
                    payload_on=field.get("payload_on", "ON"),
                    payload_off=field.get("payload_off", "OFF"),
                )

            if field["type"] == "number":
                configured = Number(
                    command_topic=f"{static_configuration['mqtt']['base_topic']}/{static_configuration['hostname']['name']}/{static_configuration['mqtt']['topic']['set']}/{field['name']}",
                    command_template=f"{ '{{ value }}' }",
                    min=field.get("min", 0),
                    max=field.get("max", 300),
                    mode=field.get("mode", "auto"),
                    step=field.get("step", 0.5),
                    optimistic=field.get("optimistic", None),
                    unit_of_measurement=field.get("unit", None),
                )

            if field["type"] == "button":
                configured = Button(
                    command_topic=f"{static_configuration['mqtt']['base_topic']}/{static_configuration['hostname']['name']}/{static_configuration['mqtt']['topic']['set']}/{field['name']}",
                    command_template=f"{ '{{ value }}' }",
                    payload_press=field.get(
                        "payload_press",
                        json.dumps(True),  # noqa: FBT003
                    ),
                )

            # Merge the configurations
            configured.__dict__ |= configuring.__dict__

            # Append the configuration to the configuration variable
            self.configuration.append(configured)

    def _update_homeassistant_configuration(self) -> None:
        """Push Home Assistant configuration for auto discovery."""
        self.logger.debug(
            "Publishing Homeassistant MQTT configuration / discovery information",
        )

        # Publish Home Assistant configuration for each field
        for field in self.configuration:
            MQTT.event.emit(
                event_name="publish",
                topic=field.configuration_topic,
                payload=json.dumps(
                    {k: v for (k, v) in field.__dict__.items() if v is not None},
                    cls=Encoder,
                ),
            )

    def _update_homeassistant_timer(self) -> None:
        """Update Home Assistant configuration update timer."""
        # Update configuration report interval timer
        self.configuration_report_timer.interval = float(
            self._static_configuration["homeassistant"]["discovery_interval"],
        )

    def _mqtt_on_connect(  # noqa: PLR0913
        self,
        client: Client,  # noqa: ARG002
        userdata: Any,  # noqa: ARG002
        flags: Any,  # noqa: ARG002
        rc: Any,  # noqa: ARG002
        reasoncode: Any = None,  # noqa: ARG002
    ) -> None:
        """On MQTT Connect, publish Home Assistant configuration."""
        self._update_homeassistant_configuration()
        self.configuration_report_timer.start()



"""Red Reactor bootstrap."""

import argparse
import logging
from importlib.metadata import version

from redreactor.components.commander import Commander
from redreactor.components.homeassistant import Homeassistant
from redreactor.components.monitor import Monitor
from redreactor.components.mqtt import MQTT
from redreactor.configuration import LinkedConfiguration


def get_arguments() -> argparse.Namespace:
    """Get passed arguments parsed."""
    parser = argparse.ArgumentParser(description="Red Reactor MQTT Client")
    parser.add_argument(
        "-c",
        "--config",
        default="config.yaml",
        help="Configuration file, defaults to `config.yaml`",
        dest="configuration_file",
    )
    parser.add_argument(
        "-d",
        "--database",
        default="database.db",
        help="Database for storing latest dynamic configuration options",
        dest="database_file",
    )
    parser.add_argument(
        "-l",
        "--log",
        default="redreactor.log",
        help="Logging file location",
        dest="logging_file",
    )

    return parser.parse_args()


def main() -> None:
    """Start Red Reactor Battery Monitor service."""
    args = get_arguments()

    # Create logger, enable all msgs
    logger = logging.getLogger("Red Reactor")
    # Set to stop INA219 logging
    logger.propagate = False
    logger.setLevel(logging.INFO)

    # Create handlers
    c_handler = logging.StreamHandler()
    f_handler = logging.FileHandler(args.logging_file)

    # Create formatters and add it to handlers
    c_format = logging.Formatter(
        "%(asctime)s - %(filename)s - %(levelname)s - %(message)s",
    )
    f_format = logging.Formatter(
        "%(asctime)s - %(filename)s - %(levelname)s - %(message)s",
        datefmt="%d-%b-%y %H:%M:%S",
    )
    c_handler.setFormatter(c_format)
    f_handler.setFormatter(f_format)

    # Add handlers to the logger
    logger.addHandler(c_handler)
    logger.addHandler(f_handler)

    logger.info("#########################################")
    logger.info("#### Red Reactor MQTT Client (%s) ####", version("redreactor"))
    logger.info("# With MQTT, and Home Assistant Support #")
    logger.info("Loading in static configuration file: %s", args.configuration_file)
    logger.info("Loading in dynamic configuration file: %s", args.database_file)

    # Load in configuration files
    configuration = LinkedConfiguration(args.configuration_file, args.database_file)

    # Update logging levels, limited to listed set
    logging_levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    if str(configuration.static["logging"]["console"]).upper() in logging_levels:
        c_handler.setLevel(configuration.static["logging"]["console"])
        logger.info(
            "Console log level set to %s",
            configuration.static["logging"]["console"],
        )
    if str(configuration.static["logging"]["file"]).upper() in logging_levels:
        f_handler.setLevel(configuration.static["logging"]["file"])
        logger.info(
            "File log level set to %s",
            configuration.static["logging"]["file"],
        )

    # Setup MQTT Connection
    mqtt = MQTT(
        configuration.static,
        exit_on_fail=bool(configuration.static["mqtt"]["exit_on_fail"]),
    )

    # Setup INA216 Battery Monitoring
    monitor = Monitor(configuration.static, configuration.dynamic)

    # Setup MQTT Commander and Battery Shutdown / Reboot commands
    Commander(configuration.static, configuration.dynamic, monitor)

    # Create Home Assistant data if enabled
    if bool(configuration.static["homeassistant"]["discovery"]):
        Homeassistant(configuration.static, configuration.dynamic)

    # Connect to MQTT server and loop forever
    mqtt.connect(
        configuration.static["mqtt"]["broker"],
        int(configuration.static["mqtt"]["port"]),
        int(configuration.static["mqtt"]["timeout"]),
    )


if __name__ == "__main__":
    main()
